#TODO flock

require 'chef/knife'
require 'erubis'

class Chef
  class Knife
    class Setup < Knife

      deps do
        require 'chef/knife/core/bootstrap_context'
        require 'chef/json_compat'
        require 'tempfile'
        require 'highline'
        require 'net/ssh'
        require 'net/ssh/multi'
        require 'net/http'
        require 'uri'
        Chef::Knife::Ssh.load_deps
      end

      banner "knife setup FQDN (options)"

      option :ssh_user,
        :short => "-x USERNAME",
        :long => "--ssh-user USERNAME",
        :description => "The ssh username",
        :default => "root"

      option :ssh_password,
        :short => "-P PASSWORD",
        :long => "--ssh-password PASSWORD",
        :description => "The ssh password"

      option :ssh_port,
        :short => "-p PORT",
        :long => "--ssh-port PORT",
        :description => "The ssh port",
        :default => "22",
        :proc => Proc.new { |key| Chef::Config[:knife][:ssh_port] = key }

      option :identity_file,
        :short => "-i IDENTITY_FILE",
        :long => "--identity-file IDENTITY_FILE",
        :description => "The SSH identity file used for authentication"

      option :chef_node_name,
        :short => "-N NAME",
        :long => "--node-name NAME",
        :description => "The Chef node name for your new node"

      option :prerelease,
        :long => "--prerelease",
        :description => "Install the pre-release chef gems"

      option :bootstrap_version,
        :long => "--bootstrap-version VERSION",
        :description => "The version of Chef to install",
        :proc => lambda { |v| Chef::Config[:knife][:bootstrap_version] = v }

      option :bootstrap_proxy,
        :long => "--bootstrap-proxy PROXY_URL",
        :description => "The proxy server for the node being bootstrapped",
        :proc => Proc.new { |p| Chef::Config[:knife][:bootstrap_proxy] = p }

      option :distro,
        :short => "-d DISTRO",
        :long => "--distro DISTRO",
        :description => "Bootstrap a distro using a template",
        :default => "debian6"

      option :use_sudo,
        :long => "--sudo",
        :description => "Execute the bootstrap via sudo",
        :boolean => true

      option :template_file,
        :long => "--template-file TEMPLATE",
        :description => "Full path to location of template to use",
        :default => false

      option :environment,
        :short => "-e ENVIRONMENT",
        :long => "--environment ENVIRONMENT",
        :description => "Set environment for node",
        :default => "_default"

      option :norunchef,
        :long => "--norunchef",
        :description => "Don't run chef after install bootstrap",
        :boolean => true,
        :default => true

      option :nobootstrap,
        :long => "--nobootstrap",
        :description => "Don't run bootstrap scripts",
        :boolean => false,
        :default => false

      option :run_list,
        :short => "-r RUN_LIST",
        :long => "--run-list RUN_LIST",
        :description => "Comma separated list of roles/recipes to apply",
        :proc => lambda { |o| o.split(/[\s,]+/) },
        :default => []

      option :no_host_key_verify,
        :long => "--no-host-key-verify",
        :description => "Disable host key verification",
        :boolean => true,
        :default => false

      def load_template(template=nil)
        # Are we bootstrapping using an already shipped template?
        if config[:template_file]
          bootstrap_files = config[:template_file]
        else
          bootstrap_files = []
          bootstrap_files << File.join(File.dirname(__FILE__), 'bootstrap', "#{config[:distro]}.erb")
          bootstrap_files << File.join(@@chef_config_dir, "bootstrap", "#{config[:distro]}.erb")
          bootstrap_files << File.join(ENV['HOME'], '.chef', 'bootstrap', "#{config[:distro]}.erb")
          bootstrap_files << Gem.find_files(File.join("chef","knife","bootstrap","#{config[:distro]}.erb"))
          bootstrap_files.flatten!
        end

        template = Array(bootstrap_files).find do |bootstrap_template|
          Chef::Log.debug("Looking for bootstrap template in #{File.dirname(bootstrap_template)}")
          File.exists?(bootstrap_template)
        end

        unless template
          ui.info("Can not find bootstrap definition for #{config[:distro]}")
          raise Errno::ENOENT
        end

        ui.info("Found bootstrap template in #{File.dirname(template)}")

        IO.read(template).chomp
      end

      def render_template(template=nil)
        context = Knife::Core::BootstrapContext.new(config, config[:run_list], Chef::Config)
        Erubis::Eruby.new(template).evaluate(context)
      end

      def get_node
        node = Chef::Node.load(server_name)
        if node.nil?
          ui.error("Could not find a node named #{@server_name}")
          exit 1
        end
        return node
      end

      def set_run_list(node, run_list)
        ui.info("Setting run_list to #{run_list.inspect} for #{node.name}")
        node.run_list.reset!(run_list)
        node.save
      end

      def set_env(node, env)
        ui.info("Setting environment to #{env} for #{node.name}")
        node.chef_environment(env)
        node.save
      end

      def bootstrap_client
        begin
          knife_ssh(ssh_command).run
        rescue Net::SSH::AuthenticationFailed
          unless config[:ssh_password]
            puts "Failed to authenticate #{config[:ssh_user]} - trying password auth"
            knife_ssh_with_password_auth(ssh_command).run
          end
        end
      end

      def run

        validate_name_args!
        @node_name = Array(@name_args).first
        # back compat--templates may use this setting:
        config[:server_name] = @node_name

        $stdout.sync = true

        ui.info("Setup Chef on #{ui.color(@node_name, :bold)}")

        bootstrap_client unless config[:nobootstrap]
        set_run_list(get_node, config[:run_list]) if config[:run_list]
        set_env(get_node, config[:environment]) unless config[:environment] == "_default"
        knife_ssh("chef-client").run if config[:norunchef]
        post if Chef::Config[:post_url]
      end

      def validate_name_args!
        if Array(@name_args).first.nil?
          ui.error("Must pass an FQDN or ip to bootstrap")
          exit 1
        end
      end

      def server_name
        config[:chef_node_name] || Array(@name_args).first
      end

      def knife_ssh(command)
        ssh = Chef::Knife::Ssh.new
        ssh.ui = ui
        ssh.name_args = [ @node_name, command]
        ssh.config[:ssh_user] = config[:ssh_user]
        ssh.config[:ssh_password] = config[:ssh_password]
        ssh.config[:ssh_port] = Chef::Config[:knife][:ssh_port] || config[:ssh_port]
        ssh.config[:identity_file] = config[:identity_file]
        ssh.config[:manual] = true
        ssh.config[:no_host_key_verify] = config[:no_host_key_verify]
        ssh.config[:on_error] = :raise
        ssh
      end

      def knife_ssh_with_password_auth(command)
        ssh = knife_ssh(command)
        ssh.config[:identity_file] = nil
        ssh.config[:ssh_password] = ssh.get_password
        ssh
      end

      def ssh_command
        command = render_template(load_template(config[:bootstrap_template]))

        if config[:use_sudo]
          command = "sudo #{command}"
        end

        command
      end

      def post
        url = Chef::Config[:post_url]
        message = {
          :fqdn    => get_node.to_hash["fqdn"]
        }
        body = JSON.generate(message)
        uri = URI.parse(url)
        request = Net::HTTP::Post.new(uri.request_uri)
        http = Net::HTTP.new(uri.host, uri.port)
        request.add_field('Content-Type', 'application/json-rpc')
        request.body = body
        response = http.request(request)
        raise "HTTP Error: #{response.code} on #{url}" unless response.code == "200" || response.code == "204"
      end

    end
  end
end